/* Copyright (c) 2009 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT. 
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 * $LastChangedRevision: 4808 $
 */ 

/** @file
@brief Implementation of the ACI transport layer module
*/

#include "hal_platform.h"
#include "hal_aci_tl.h"
#include "hal_io.h"
#include "system.h"
#include "assertions.h"
#include "lib_mem.h"
#include <string.h>

#define ACI_QUEUE_SIZE  16

typedef struct {
 hal_aci_data_t           aci_data[ACI_QUEUE_SIZE];
 uint8_t                  head;
 uint8_t                  tail;
} aci_queue_t;

static hal_aci_data_t received_data;
static uint8_t spi_readwrite(uint8_t aci_byte);
static aci_queue_t aci_tx_q;
static aci_queue_t aci_rx_q;

static void m_aci_q_init(aci_queue_t *aci_q)
{
  aci_q->head = 0;
  aci_q->tail = 0;
}

static bool m_aci_q_enqueue(aci_queue_t *aci_q, hal_aci_data_t *p_data)
{
  const uint8_t next = (aci_q->tail + 1) % ACI_QUEUE_SIZE;
  const uint8_t length = p_data->buffer[0];

  if (next == aci_q->head)
  {
    /* full queue */
    return false;
  }
  aci_q->aci_data[aci_q->tail].status_byte = 0;
  // @comment : It would be better to copy the entire hal_aci_data_t into the buffer.
  memcpy((uint8_t *)&(aci_q->aci_data[aci_q->tail].buffer[0]), (uint8_t *)&p_data->buffer[0], length + 1);
  aci_q->tail = next;

  return true;
}

//@comment have a unit test for the queue, esp. the full and the empty states
static bool m_aci_q_dequeue(aci_queue_t *aci_q, hal_aci_data_t *p_data)
{
  if (aci_q->head == aci_q->tail)
  {
    /* empty queue */
    return false;
  }

  memcpy((uint8_t *)p_data, (uint8_t *)&(aci_q->aci_data[aci_q->head]), sizeof(hal_aci_data_t));
  aci_q->head = (aci_q->head + 1) % ACI_QUEUE_SIZE;

  return true;
}

static bool m_aci_q_is_full(aci_queue_t *aci_q)
{
  uint8_t next;
  bool state;

  //This should be done in a critical section
  DISABLE_INTERRUPTS();
  next = (aci_q->tail + 1) % ACI_QUEUE_SIZE;

  if (next == aci_q->head)
  {
    state = true;
  }
  else
  {
    state = false;
  }

  ENABLE_INTERRUPTS();
  //end

  return state;
}

void EINT3_IRQHandler(void)
{
  hal_aci_data_t *p_aci_data;

  if (LPC_GPIOINT->IO0IntStatF == (1UL << 16UL))
  {
    LPC_GPIOINT->IO0IntClr = (1UL << 16UL);

    // Receive or transmit data
    p_aci_data = hal_aci_tl_poll_get();

    // Check if we received data
    if (p_aci_data->buffer[0] > 0)
    {
      if (!m_aci_q_enqueue(&aci_rx_q, p_aci_data))
      {
        /* Receive Buffer full.
           Should never happen.
           */
         ASSERT(ERROR_CODE_HAL_ACI_TL_OVERFLOW, false);
      }
      if (m_aci_q_is_full(&aci_rx_q))
      {
        /* Disable RDY line interrupt.
           Will latch any pending RDY lines, so when enabled this
           routine should be taken again */
        hal_pltf_disable_ready_interrupt();
      }
    }
  }
}

bool hal_aci_tl_event_get(hal_aci_data_t *p_aci_data)
{
  bool was_full = m_aci_q_is_full(&aci_rx_q);

  if (m_aci_q_dequeue(&aci_rx_q, p_aci_data))
  {
    if (was_full)
    {
      /* Enable RDY line interrupt again */
      hal_pltf_enable_ready_interrupt();
    }
    return true;
  }
  else
  {
    return false;
  }
}

void hal_aci_tl_io_config()
{
  int32_t tmp;

  HAL_IO_CONFIG(HAL_IO_RADIO_REQN, HAL_IO_OUTPUT, HAL_IO_OUTPUT_NORMAL_STRENGTH);
  HAL_IO_SET_STATE(HAL_IO_RADIO_REQN, 1);
  HAL_IO_CONFIG(HAL_IO_RADIO_SCK, HAL_IO_OUTPUT, HAL_IO_OUTPUT_NORMAL_STRENGTH);
  HAL_IO_SET_STATE(HAL_IO_RADIO_SCK, 0);
  HAL_IO_CONFIG(HAL_IO_RADIO_MOSI, HAL_IO_OUTPUT, HAL_IO_OUTPUT_NORMAL_STRENGTH);
  
  HAL_IO_CONFIG(HAL_IO_RADIO_ACTIVE, HAL_IO_INPUT, HAL_IO_INPUT_BUF_ON_NO_PULL);
  HAL_IO_CONFIG(HAL_IO_RADIO_RDY, HAL_IO_INPUT, HAL_IO_INPUT_BUF_ON_PULL_UP);
  HAL_IO_CONFIG(HAL_IO_RADIO_MISO, HAL_IO_INPUT, HAL_IO_INPUT_BUF_ON_PULL_UP);
  
  HAL_IO_CONFIG(HAL_IO_RADIO_RST, HAL_IO_OUTPUT, HAL_IO_INPUT_BUF_ON_PULL_UP);
  HAL_IO_SET_STATE(HAL_IO_RADIO_RST, 0);
  tmp = 30000;
  while (tmp--) ;
  HAL_IO_SET_STATE(HAL_IO_RADIO_RST, 1);
}

void hal_aci_tl_init()
{
  hal_aci_tl_io_config();
  received_data.buffer[0] = 0;

  /* initialize aci cmd queue */
  m_aci_q_init(&aci_tx_q);
  m_aci_q_init(&aci_rx_q);

  hal_pltf_configure_spi_for_aci();
  hal_pltf_disable_spi();  
  hal_pltf_config_ready_line();
  hal_pltf_enable_ready_interrupt();
  HAL_IO_SET_STATE(HAL_IO_RADIO_REQN, 0);
  HAL_IO_SET_STATE(HAL_IO_RADIO_REQN, 1);
}

bool hal_aci_tl_send(hal_aci_data_t *p_aci_cmd)
{
  const uint8_t length = p_aci_cmd->buffer[0];
  bool ret_val = false;

  //ASSERT(ERROR_CODE_HAL_ACI_TL_OVERFLOW,(p_aci_cmd->buffer[0] <= HAL_ACI_MAX_LENGTH));
  if (length > HAL_ACI_MAX_LENGTH)
  {
    return false;
  }
  else
  {
    if (m_aci_q_enqueue(&aci_tx_q, p_aci_cmd))
    {
      ret_val = true;
    }
  }
  HAL_IO_SET_STATE(HAL_IO_RADIO_REQN, 0);
  return ret_val;
}

hal_aci_data_t * hal_aci_tl_poll_get(void)
{
  uint8_t byte_cnt;
  uint8_t byte_sent_cnt;
  uint8_t max_bytes;
  hal_aci_data_t data_to_send;

  hal_pltf_enable_spi();

  HAL_IO_SET_STATE(HAL_IO_RADIO_REQN, 0);

  // Receive from queue
  if (m_aci_q_dequeue(&aci_tx_q, &data_to_send) == false)
  {
    /* queue was empty, nothing to send */
    data_to_send.status_byte = 0;
    data_to_send.buffer[0] = 0;
  }

  //Change this if your mcu has DMA for the master SPI

  // Send length, receive header
  byte_sent_cnt = 0;
  received_data.status_byte = spi_readwrite(data_to_send.buffer[byte_sent_cnt++]);
  // Send first byte, receive length from slave
  received_data.buffer[0] = spi_readwrite(data_to_send.buffer[byte_sent_cnt++]);
  if (0 == data_to_send.buffer[0])
  {
    max_bytes = received_data.buffer[0];
  }
  else
  {
    // Set the maximum to the biggest size. One command byte is already sent
    max_bytes = (received_data.buffer[0] > (data_to_send.buffer[0] - 1))
      ? received_data.buffer[0] : (data_to_send.buffer[0] - 1);
  }

  if (max_bytes > HAL_ACI_MAX_LENGTH)
  {
    max_bytes = HAL_ACI_MAX_LENGTH;
  }

  // Transmit/receive the rest of the packet
  for (byte_cnt = 0; byte_cnt < max_bytes; byte_cnt++)
  {
    received_data.buffer[byte_cnt+1] =  spi_readwrite(data_to_send.buffer[byte_sent_cnt++]);
  }

  HAL_IO_SET_STATE(HAL_IO_RADIO_REQN, 1);

  hal_pltf_disable_spi();

  ASSERT(ERROR_CODE_HAL_ACI_TL_STATUS_BYTE,(0 != received_data.status_byte));

  /* valid Rx available or transmit finished*/
  return (&received_data);
}

static uint8_t spi_readwrite(const uint8_t aci_byte)
{
  hal_pltf_set_spim_data(aci_byte);      
  while(!hal_pltf_is_spim_data_ready())
    ;
  return hal_pltf_get_spim_data();
}

