/* Copyright (c) 2010 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 * $LastChangedRevision:  $
 */

/** @file
 * @brief ACI transport layer demo source.
 */

/** @defgroup aci_tl_demo aci_tl_demo
@ingroup projects
@brief Simple ACI transport layer demo.
@details

@section aci_tl_demo_intro Introduction

The aci_tl_demo performs a simple test of hal_aci_tl:
 -# Sets nRF8001 in Test state.
 -# Sends one @c Echo command.
 -# Receives the @c EchoEvent.
 -# Compare the data from the @c EchoEvent with the contents in the @c Echo command.
 -# Displays the results.

@section aci_tl_demo_setup I/O setup and interpretation

The aci_tl_demo uses two LEDs to display the results:
 - LED0: Port pin P10
 - LED1: Port pin P11

When the firmware starts, both LEDs are off. When finished you should see one of the
following results:
 - Both LEDs are lit: The test passed.
 - Only LED0 is lit: An error occured.

Several checks are performed during the execution of the firmware. A failing check will
cause the error message to be displayed. The checks are (in order):
 -# A @c DeviceStartedEvent is received with state set to Standby when starting.
 -# A @c DeviceStartedEvent is received with state set to Test in reply to the @c Test
    command.
 -# An @c EchoEvent is received in reply to @c Echo command with:
    - Correct length.
    - Identical content to the @c Echo message.


*/

#ifdef __USE_CMSIS
#include "LPC17xx.h"
#endif

#include <cr_section_macros.h>
#include <NXP/crp.h>
#include "hal_platform.h"
#include "hal_aci_tl.h"
#include "system.h"
#include "hal_io.h"
#include "lib_mem.h"

// Variable to store CRP value in. Will be placed automatically
// by the linker when "Enable Code Read Protect" selected.
// See crp.h header for more information
__CRP const unsigned int CRP_WORD = CRP_NO_CRP ;

/*ACI commands opcode*/
#define OPCODE_DEVICE_SET_TEST_MODE_CMD  0x01
#define OPCODE_DEVICE_ECHO               0x02

/*ACI commands parameters*/
#define TEST_MODE_DTM_ACI                0x02

/*commands messages to be sent to nRF8001*/
hal_aci_data_t cmd_set_mode_test   = {0, {2, OPCODE_DEVICE_SET_TEST_MODE_CMD, TEST_MODE_DTM_ACI}};

/*dummy message to send/receive in loopback mode*/
hal_aci_data_t dummy_msg_msg = {0, {29, OPCODE_DEVICE_ECHO, 0xaa, 0xff, 0x43, 0xe1, 0xFE, 0xDE, 0x08, 0x09, 0x0a,
                                      0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14,
                                      0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e,
                                      }};


/*ACI events opcode*/
#define OPCODE_DEVICE_STARTED_EVT       0x81
#define OPCODE_CMD_RESPONSE_EVT         0x84
#define OPCODE_ECHO_EVT                 0x82

/*ACI events parameters*/
#define MODE_SETUP                      0x02
#define MODE_TEST                       0x01
#define MODE_STDBY                      0x03

#define CMD_SUCCESS                     0x00

/*global variables*/
static hal_aci_data_t last_rcvd_msg;

void error_handler()
{
  do
  {
    HAL_IO_SET_STATE(HAL_IO_LED0, 1);
  }while(1);
}


int main(void)
{
  uint8_t index;

  hal_aci_tl_io_config();
  hal_aci_tl_init();
  ENABLE_INTERRUPTS();

  /*wait for event device_started in stdby mode*/
  while(!hal_aci_tl_event_get(&last_rcvd_msg)) ;

  /*check if received message is the expected device started event*/
  if((OPCODE_DEVICE_STARTED_EVT != last_rcvd_msg.buffer[1])||(MODE_SETUP != last_rcvd_msg.buffer[2]))
  {
    error_handler();
  }

  /*send set_test_mode command to enter test_mode*/
  cmd_set_mode_test.buffer[0] = 2;
  hal_aci_tl_send(&cmd_set_mode_test);

  /*wait for event device_started in test mode*/
  while(!hal_aci_tl_event_get(&last_rcvd_msg)) ;

  /*check if received message is the expected device started event*/
  if((OPCODE_DEVICE_STARTED_EVT != last_rcvd_msg.buffer[1])||(MODE_TEST != last_rcvd_msg.buffer[2]))
  {
   error_handler();
  }

  /*send dummy message (with command byte set to echo command (0x0E)*/
  hal_aci_tl_send(&dummy_msg_msg);

  /*wait for the dummy message to come back*/
  while(!hal_aci_tl_event_get(&last_rcvd_msg)) ;

  /*check if received message is the dummy message*/
  if (dummy_msg_msg.buffer[0] == last_rcvd_msg.buffer[0])   //check that length is correct
  {
    if (OPCODE_ECHO_EVT != last_rcvd_msg.buffer[1])         //check that the event is an echo event
    {
      error_handler();
    }
    else
    {
      for (index = 2; index < dummy_msg_msg.buffer[0]; index ++)
      {
        if (last_rcvd_msg.buffer[index] != dummy_msg_msg.buffer[index])
        {
          error_handler();
        }
      }
    }
  }
  else
  {
    error_handler();
  }


  do
  {
    HAL_IO_SET_STATE(HAL_IO_LED0, 1);
    HAL_IO_SET_STATE(HAL_IO_LED1, 1);
  } while(1);

  return 0;
}

void SystemInit(void)
{
}

